from collections import namedtuple
import errno
import os
import shutil
import warnings

import click
import logging
import pandas as pd
from attr import dataclass
from zipline.data.adjustments import SQLiteAdjustmentReader, SQLiteAdjustmentWriter
from zipline.utils.calendar_utils import get_calendar
from toolz import curry, complement, take

from zipline.assets import AssetDBWriter, AssetFinder, ASSET_DB_VERSION, Asset
from zipline.assets.asset_db_migrations import downgrade
from zipline.utils.cache import (
    dataframe_cache,
    working_dir,
    working_file,
)
from zipline.utils.compat import ExitStack, mappingproxy
from zipline.utils.input_validation import ensure_timestamp, optionally
import zipline.utils.paths as pth
from zipline.utils.preprocess import preprocess

from ziplime.constants.default_columns import DEFAULT_COLUMNS
from ziplime.data.abstract_data_bundle import AbstractDataBundle
from ziplime.data.abstract_fundamendal_data_provider import AbstractFundamentalDataProvider
from ziplime.data.abstract_historical_market_data_provider import AbstractHistoricalMarketDataProvider
from ziplime.data.bcolz_daily_bars import ZiplimeBcolzDailyBarWriter
from ziplime.data.bcolz_minute_bars import ZiplimeBcolzMinuteBarWriter, ZiplimeBcolzMinuteBarReader
from ziplime.data.storages.bcolz_data_bundle import BcolzDataBundle
from ziplime.domain.column_specification import ColumnSpecification

log = logging.getLogger(__name__)


def asset_db_path(bundle_name, timestr, environ=None, db_version=None):
    return pth.data_path(
        asset_db_relative(bundle_name, timestr, db_version),
        environ=environ,
    )


def minute_equity_path(bundle_name, timestr, environ=None):
    return pth.data_path(
        minute_equity_relative(bundle_name, timestr),
        environ=environ,
    )


def daily_equity_path(bundle_name, timestr, environ=None):
    return pth.data_path(
        daily_equity_relative(bundle_name, timestr),
        environ=environ,
    )


def fundamental_data_path(bundle_name, timestr, environ=None):
    return pth.data_path(
        fundamental_data_relative(bundle_name, timestr),
        environ=environ,
    )


def adjustment_db_path(bundle_name, timestr, environ=None):
    return pth.data_path(
        adjustment_db_relative(bundle_name, timestr),
        environ=environ,
    )


def cache_path(bundle_name, environ=None):
    return pth.data_path(
        cache_relative(bundle_name),
        environ=environ,
    )


def adjustment_db_relative(bundle_name, timestr):
    return bundle_name, timestr, "adjustments.sqlite"


def cache_relative(bundle_name):
    return bundle_name, ".cache"


def daily_equity_relative(bundle_name, timestr):
    return bundle_name, timestr, "daily_equities.bcolz"


def fundamental_data_relative(bundle_name, timestr):
    return bundle_name, timestr, "fundamental_data.bcolz"


def minute_equity_relative(bundle_name, timestr):
    return bundle_name, timestr, "minute_equities.bcolz"


def asset_db_relative(bundle_name, timestr, db_version=None):
    db_version = ASSET_DB_VERSION if db_version is None else db_version

    return bundle_name, timestr, "assets-%d.sqlite" % db_version


def to_bundle_ingest_dirname(ts):
    """Convert a pandas Timestamp into the name of the directory for the
    ingestion.

    Parameters
    ----------
    ts : pandas.Timestamp
        The time of the ingestions

    Returns
    -------
    name : str
        The name of the directory for this ingestion.
    """
    return ts.isoformat().replace(":", ";")


def from_bundle_ingest_dirname(cs):
    """Read a bundle ingestion directory name into a pandas Timestamp.

    Parameters
    ----------
    cs : str
        The name of the directory.

    Returns
    -------
    ts : pandas.Timestamp
        The time when this ingestion happened.
    """
    return pd.Timestamp(cs.replace(";", ":"))


def ingestions_for_bundle(bundle, environ=None):
    return sorted(
        (
            from_bundle_ingest_dirname(ing)
            for ing in os.listdir(pth.data_path([bundle], environ))
            if not pth.hidden(ing)
        ),
        reverse=True,
    )


RegisteredBundle = namedtuple(
    "RegisteredBundle",
    [
        "calendar_name",
        "start_session",
        "end_session",
        "minutes_per_day",
        "ingest",
        "create_writers",
    ],
)


@dataclass
class BundleData:
    name: str
    asset_finder: AssetFinder
    equity_minute_bar_reader: ZiplimeBcolzMinuteBarReader
    equity_daily_bar_reader: AbstractDataBundle
    fundamental_data_reader: AbstractDataBundle

    adjustment_reader: SQLiteAdjustmentReader

    cached_data: pd.DataFrame = None


# BundleData = namedtuple(
#     "BundleData",
#     "asset_finder equity_minute_bar_reader equity_daily_bar_reader "
#     "adjustment_reader",
# )

BundleCore = namedtuple(
    "BundleCore",
    "bundles register unregister ingest load clean",
)


class UnknownBundle(click.ClickException, LookupError):
    """Raised if no bundle with the given name was registered."""

    exit_code = 1

    def __init__(self, name):
        super(UnknownBundle, self).__init__(
            "No bundle registered with the name %r" % name,
        )
        self.name = name

    def __str__(self):
        return self.message


class BadClean(click.ClickException, ValueError):
    """Exception indicating that an invalid argument set was passed to
    ``clean``.

    Parameters
    ----------
    before, after, keep_last : any
        The bad arguments to ``clean``.

    See Also
    --------
    clean
    """

    def __init__(self, before, after, keep_last):
        super(BadClean, self).__init__(
            "Cannot pass a combination of `before` and `after` with "
            "`keep_last`. Must pass one. "
            "Got: before=%r, after=%r, keep_last=%r\n"
            % (
                before,
                after,
                keep_last,
            ),
        )

    def __str__(self):
        return self.message


# TODO: simplify
# flake8: noqa: C901
def _make_bundle_core():
    """Create a family of data bundle functions that read from the same
    bundle mapping.

    Returns
    -------
    bundles : mappingproxy
        The mapping of bundles to bundle payloads.
    register : callable
        The function which registers new bundles in the ``bundles`` mapping.
    unregister : callable
        The function which deregisters bundles from the ``bundles`` mapping.
    ingest : callable
        The function which downloads and write data for a given data bundle.
    load : callable
        The function which loads the ingested bundles back into memory.
    clean : callable
        The function which cleans up data written with ``ingest``.
    """
    _bundles = {}  # the registered bundles
    # Expose _bundles through a proxy so that users cannot mutate this
    # accidentally. Users may go through `register` to update this which will
    # warn when trampling another bundle.
    bundles = mappingproxy(_bundles)

    @curry
    def register(
            name,
            f,
            calendar_name="NYSE",
            start_session=None,
            end_session=None,
            minutes_per_day=390,
            create_writers=True,
    ):
        """Register a data bundle ingest function.

        Parameters
        ----------
        name : str
            The name of the bundle.
        f : callable
            The ingest function. This function will be passed:

              environ : mapping
                  The environment this is being run with.
              asset_db_writer : AssetDBWriter
                  The asset db writer to write into.
              minute_bar_writer : BcolzMinuteBarWriter
                  The minute bar writer to write into.
              daily_bar_writer : BcolzDailyBarWriter
                  The daily bar writer to write into.
              adjustment_writer : SQLiteAdjustmentWriter
                  The adjustment db writer to write into.
              calendar : trading_calendars.TradingCalendar
                  The trading calendar to ingest for.
              start_session : pd.Timestamp
                  The first session of data to ingest.
              end_session : pd.Timestamp
                  The last session of data to ingest.
              cache : DataFrameCache
                  A mapping object to temporarily store dataframes.
                  This should be used to cache intermediates in case the load
                  fails. This will be automatically cleaned up after a
                  successful load.
              show_progress : bool
                  Show the progress for the current load where possible.
        calendar_name : str, optional
            The name of a calendar used to align bundle data.
            Default is 'NYSE'.
        start_session : pd.Timestamp, optional
            The first session for which we want data. If not provided,
            or if the date lies outside the range supported by the
            calendar, the first_session of the calendar is used.
        end_session : pd.Timestamp, optional
            The last session for which we want data. If not provided,
            or if the date lies outside the range supported by the
            calendar, the last_session of the calendar is used.
        minutes_per_day : int, optional
            The number of minutes in each normal trading day.
        create_writers : bool, optional
            Should the ingest machinery create the writers for the ingest
            function. This can be disabled as an optimization for cases where
            they are not needed, like the ``quantopian-quandl`` bundle.

        Notes
        -----
        This function my be used as a decorator, for example:

        .. code-block:: python

           @register('quandl')
           def quandl_ingest_function(...):
               ...

        See Also
        --------
        zipline.data.bundles.bundles
        """
        # if name in bundles:
        #     warnings.warn(
        #         "Overwriting bundle with name %r" % name,
        #         stacklevel=3,
        #     )

        # NOTE: We don't eagerly compute calendar values here because
        # `register` is called at module scope in zipline, and creating a
        # calendar currently takes between 0.5 and 1 seconds, which causes a
        # noticeable delay on the zipline CLI.
        _bundles[name] = RegisteredBundle(
            calendar_name=calendar_name,
            start_session=start_session,
            end_session=end_session,
            minutes_per_day=minutes_per_day,
            ingest=f,
            create_writers=create_writers,
        )
        return f

    def unregister(name):
        """Unregister a bundle.

        Parameters
        ----------
        name : str
            The name of the bundle to unregister.

        Raises
        ------
        UnknownBundle
            Raised when no bundle has been registered with the given name.

        See Also
        --------
        zipline.data.bundles.bundles
        """
        try:
            del _bundles[name]
        except KeyError:
            raise UnknownBundle(name)

    def ingest(
            name,
            fundamental_data_provider: AbstractFundamentalDataProvider,
            historical_market_data_provider: AbstractHistoricalMarketDataProvider,
            daily_bar_writer_class: AbstractDataBundle,
            fundamental_data_writer_class: AbstractDataBundle,
            market_data_fields: list[ColumnSpecification],
            fundamental_data_fields: list[ColumnSpecification],
            environ=os.environ,
            timestamp=None,
            assets_versions=(),
            show_progress: bool = False,
            minute_bar_writer_class=ZiplimeBcolzMinuteBarWriter,
            minute_bar_writer_cols: list[ColumnSpecification] = DEFAULT_COLUMNS,
            **kwargs,
    ):
        """Ingest data for a given bundle.

        Parameters
        ----------
        name : str
            The name of the bundle.
        environ : mapping, optional
            The environment variables. By default this is os.environ.
        timestamp : datetime, optional
            The timestamp to use for the load.
            By default this is the current time.
        assets_versions : Iterable[int], optional
            Versions of the assets db to which to downgrade.
        show_progress : bool, optional
            Tell the ingest function to display the progress where possible.
        kwargs : Any, optional
            Additional kwargs being passed to the ingest function.
        """
        try:
            bundle = bundles[name]
        except KeyError:
            raise UnknownBundle(name)

        calendar = get_calendar(bundle.calendar_name)

        start_session = bundle.start_session
        end_session = bundle.end_session

        if start_session is None or start_session < calendar.first_session:
            start_session = calendar.first_session

        if end_session is None or end_session > calendar.last_session:
            end_session = calendar.last_session

        if timestamp is None:
            timestamp = pd.Timestamp.utcnow()
        timestamp = timestamp.tz_convert("utc").tz_localize(None)

        timestr = to_bundle_ingest_dirname(timestamp)
        cachepath = cache_path(name, environ=environ)
        pth.ensure_directory(pth.data_path([name, timestr], environ=environ))
        pth.ensure_directory(cachepath)
        with dataframe_cache(
                cachepath, clean_on_failure=False
        ) as cache, ExitStack() as stack:
            # we use `cleanup_on_failure=False` so that we don't purge the
            # cache directory if the load fails in the middle
            if bundle.create_writers:
                wd = stack.enter_context(
                    working_dir(pth.data_path([], environ=environ))
                )
                daily_bars_path = wd.ensure_dir(*daily_equity_relative(name, timestr))
                daily_bar_writer = daily_bar_writer_class(
                    daily_bars_path
                )

                fundamental_data_path = wd.ensure_dir(*fundamental_data_relative(name, timestr))
                fundamental_data_writer = fundamental_data_writer_class(
                    fundamental_data_path
                )
                # Do an empty write to ensure that the daily ctables exist
                # when we create the SQLiteAdjustmentWriter below. The
                # SQLiteAdjustmentWriter needs to open the daily ctables so
                # that it can compute the adjustment ratios for the dividends.

                minute_bar_writer = minute_bar_writer_class(
                    wd.ensure_dir(*minute_equity_relative(name, timestr)),
                    calendar,
                    start_session,
                    end_session,
                    cols=minute_bar_writer_cols,
                    minutes_per_day=bundle.minutes_per_day,
                )
                assets_db_path = wd.getpath(*asset_db_relative(name, timestr))
                asset_db_writer = AssetDBWriter(assets_db_path)

                adjustment_db_writer = stack.enter_context(
                    SQLiteAdjustmentWriter(
                        wd.getpath(*adjustment_db_relative(name, timestr)),
                        daily_bar_writer_class(daily_bars_path),
                        overwrite=True,
                    )
                )
            else:
                daily_bar_writer = None
                minute_bar_writer = None
                asset_db_writer = None
                adjustment_db_writer = None
                if assets_versions:
                    raise ValueError(
                        "Need to ingest a bundle that creates "
                        "writers in order to downgrade the assets"
                        " db."
                    )
            log.info("Ingesting %s", name)
            bundle.ingest(
                environ=environ,
                historical_market_data_provider=historical_market_data_provider,
                fundamental_data_provider=fundamental_data_provider,
                asset_db_writer=asset_db_writer,
                minute_bar_writer=minute_bar_writer,
                daily_bar_writer=daily_bar_writer,
                fundamental_data_writer=fundamental_data_writer,
                adjustment_writer=adjustment_db_writer,
                calendar=calendar,
                start_session=start_session,
                end_session=end_session,
                cache=cache,
                show_progress=show_progress,
                output_dir=pth.data_path([name, timestr], environ=environ),
                market_data_fields=market_data_fields,
                fundamental_data_fields=fundamental_data_fields,
                **kwargs,
            )

            for version in sorted(set(assets_versions), reverse=True):
                version_path = wd.getpath(
                    *asset_db_relative(
                        name,
                        timestr,
                        db_version=version,
                    )
                )
                with working_file(version_path) as wf:
                    shutil.copy2(assets_db_path, wf.path)
                    downgrade(wf.path, version)

    def most_recent_data(bundle_name, timestamp, environ=None):
        """Get the path to the most recent data after ``date``for the
        given bundle.

        Parameters
        ----------
        bundle_name : str
            The name of the bundle to lookup.
        timestamp : datetime
            The timestamp to begin searching on or before.
        environ : dict, optional
            An environment dict to forward to zipline_root.
        """
        if bundle_name not in bundles:
            raise UnknownBundle(bundle_name)

        try:
            candidates = os.listdir(
                pth.data_path([bundle_name], environ=environ),
            )
            return pth.data_path(
                [
                    bundle_name,
                    max(
                        filter(complement(pth.hidden), candidates),
                        key=from_bundle_ingest_dirname,
                    ),
                ],
                environ=environ,
            )
        except (ValueError, OSError) as e:
            if getattr(e, "errno", errno.ENOENT) != errno.ENOENT:
                raise
            raise ValueError(
                "no data for bundle {bundle!r} on or before {timestamp}\n"
                "maybe you need to run: $ zipline ingest -b {bundle}".format(
                    bundle=bundle_name,
                    timestamp=timestamp,
                ),
            )

    def load(name, environ=os.environ, timestamp=None):
        """Loads a previously ingested bundle.

        Parameters
        ----------
        name : str
            The name of the bundle.
        environ : mapping, optional
            The environment variables. Defaults of os.environ.
        timestamp : datetime, optional
            The timestamp of the data to lookup.
            Defaults to the current time.

        Returns
        -------
        bundle_data : BundleData
            The raw data readers for this bundle.
        """
        if timestamp is None:
            timestamp = pd.Timestamp.utcnow()
        timestr = most_recent_data(name, timestamp, environ=environ)
        return BundleData(
            name=name,
            asset_finder=AssetFinder(
                asset_db_path(name, timestr, environ=environ),
            ),
            equity_minute_bar_reader=ZiplimeBcolzMinuteBarReader(
                minute_equity_path(name, timestr, environ=environ),
            ),
            equity_daily_bar_reader=BcolzDataBundle(
                root_directory=daily_equity_path(name, timestr, environ=environ),
            ),
            fundamental_data_reader=BcolzDataBundle(
                root_directory=fundamental_data_path(name, timestr, environ=environ),
            ),

            # equity_daily_bar_reader=ZiplimeBcolzDailyBarReader(
            #     daily_equity_path(name, timestr, environ=environ),
            # ),
            adjustment_reader=SQLiteAdjustmentReader(
                adjustment_db_path(name, timestr, environ=environ),
            ),
        )

    @preprocess(
        before=optionally(ensure_timestamp),
        after=optionally(ensure_timestamp),
    )
    def clean(name, before=None, after=None, keep_last=None, environ=os.environ):
        """Clean up data that was created with ``ingest`` or
        ``$ python -m zipline ingest``

        Parameters
        ----------
        name : str
            The name of the bundle to remove data for.
        before : datetime, optional
            Remove data ingested before this date.
            This argument is mutually exclusive with: keep_last
        after : datetime, optional
            Remove data ingested after this date.
            This argument is mutually exclusive with: keep_last
        keep_last : int, optional
            Remove all but the last ``keep_last`` ingestions.
            This argument is mutually exclusive with:
              before
              after
        environ : mapping, optional
            The environment variables. Defaults of os.environ.

        Returns
        -------
        cleaned : set[str]
            The names of the runs that were removed.

        Raises
        ------
        BadClean
            Raised when ``before`` and or ``after`` are passed with
            ``keep_last``. This is a subclass of ``ValueError``.
        """
        try:
            all_runs = sorted(
                filter(
                    complement(pth.hidden),
                    os.listdir(pth.data_path([name], environ=environ)),
                ),
                key=from_bundle_ingest_dirname,
            )
        except OSError as e:
            if e.errno != errno.ENOENT:
                raise
            raise UnknownBundle(name)

        if before is after is keep_last is None:
            raise BadClean(before, after, keep_last)
        if (before is not None or after is not None) and keep_last is not None:
            raise BadClean(before, after, keep_last)

        if keep_last is None:

            def should_clean(name):
                dt = from_bundle_ingest_dirname(name)
                return (before is not None and dt < before) or (
                        after is not None and dt > after
                )

        elif keep_last >= 0:
            last_n_dts = set(take(keep_last, reversed(all_runs)))

            def should_clean(name):
                return name not in last_n_dts

        else:
            raise BadClean(before, after, keep_last)

        cleaned = set()
        for run in all_runs:
            if should_clean(run):
                log.info("Cleaning %s.", run)
                path = pth.data_path([name, run], environ=environ)
                shutil.rmtree(path)
                cleaned.add(path)

        return cleaned

    return BundleCore(bundles, register, unregister, ingest, load, clean)


bundles, register, unregister, ingest, load, clean = _make_bundle_core()



import datetime

import pandas as pd
from exchange_calendars import get_calendar


def normalize_daily_start_end_session(
        calendar_name: str, start_session: datetime.datetime,
        end_session: datetime.datetime) -> tuple[pd.Timestamp, pd.Timestamp]:
    cal = get_calendar(calendar_name, start=start_session)
    if start_session < cal.first_session:  # eg. Starts on 1 Jan will be realigned to the first trading day of the year
        start_session = cal.first_session
        if start_session.weekday() == 6:  # Don't start on Sundays, this helps with futures testing...
            start_session = cal.next_close(start_session).floor(freq="D")
    if not (cal.is_session(end_session)):
        end_session = cal.previous_close(end_session).floor(freq="D")

    start_session = pd.Timestamp(ts_input=start_session, tzinfo=None).tz_localize(None)
    end_session = pd.Timestamp(ts_input=end_session, tzinfo=None).tz_localize(None)
    return start_session, end_session



import datetime
import logging
import multiprocessing
from dataclasses import asdict
from queue import Queue

import pandas as pd
from joblib import Parallel, delayed
from lime_trader import LimeClient
from lime_trader.models.market import Period

from ziplime.data.abstract_live_market_data_provider import AbstractLiveMarketDataProvider
from ziplime.domain.lime_quote import LimeQuote


class LimeTraderSdkLiveMarketDataProvider(AbstractLiveMarketDataProvider):
    def __init__(self, lime_sdk_credentials_file: str):
        self._lime_sdk_credentials_file = lime_sdk_credentials_file
        self._logger = logging.getLogger(__name__)
        self._lime_sdk_client = LimeClient.from_file(lime_sdk_credentials_file, logger=self._logger)

    def fetch_live_data_table(
            self,
            symbols: list[str],
            period: Period,
            date_from: datetime.datetime,
            date_to: datetime.datetime,
            show_progress: bool):

        live_data_queue = Queue()

        def fetch_live(lime_trader_sdk_credentials_file: str, symbol: str):
            lime_client = LimeClient.from_file(lime_trader_sdk_credentials_file, logger=self._logger)
            try:
                quotes = lime_client.market.get_quotes_history(
                    symbol=symbol, period=period, from_date=date_from,
                    to_date=date_to
                )
                df = LimeTraderSdkLiveMarketDataProvider.load_data_table(
                    quotes=[LimeQuote(symbol=symbol, quote_history=quote) for quote in quotes],
                    show_progress=show_progress
                )
            except Exception as e:
                self._logger.error("Error fetching data using lime trader sdk")
                df = pd.DataFrame()
            live_data_queue.put(df)

        res = Parallel(n_jobs=multiprocessing.cpu_count() * 2, prefer="threads", return_as="generator_unordered", )(
            delayed(fetch_live)(self._lime_sdk_credentials_file, symbol) for symbol in symbols)

        if show_progress:
            self._logger.info("Downloading live Lime Trader SDK metadata.")
        processed_symbols = 0
        while True:
            if processed_symbols == len(symbols):
                break
            item = live_data_queue.get()
            yield item
            processed_symbols += 1

    @staticmethod
    def load_data_table(quotes: list[LimeQuote], show_progress: bool = False):
        if not quotes:
            return pd.DataFrame()
        data_table = pd.DataFrame(
            [dict(**asdict(quote_hist.quote_history), symbol=quote_hist.symbol) for quote_hist in quotes], )

        data_table.rename(
            columns={
                "timestamp": "date",
            },
            inplace=True,
            copy=False,
        )
        # data_table = data_table.reset_index()
        data_table = data_table.set_index('date', drop=False)
        return data_table



import datetime

from lime_trader.models.market import Period

from ziplime.constants.fundamental_data import FUNDAMENTAL_DATA_COLUMNS, FUNDAMENTAL_DATA_COLUMN_NAMES
from ziplime.data.bundles.lime import register_lime_equities_bundle


def register_lime_symbol_list_equities_bundle(bundle_name: str,
                                              start_session: datetime.datetime | None,
                                              end_session: datetime.datetime | None,
                                              symbols: list[str],
                                              fundamental_data_list: set[str] = FUNDAMENTAL_DATA_COLUMN_NAMES,
                                              period: Period = Period.DAY,
                                              calendar_name: str = "NYSE"
                                              ):
    register_lime_equities_bundle(
        bundle_name=bundle_name,
        start_session=start_session,
        end_session=end_session,
        symbol_list=symbols,
        period=period,
        calendar_name=calendar_name,
        fundamental_data_list=fundamental_data_list,
    )



import datetime
from abc import abstractmethod
from lime_trader.models.market import Period


class AbstractHistoricalMarketDataProvider:

    @abstractmethod
    def get_historical_data_table(self, symbols: list[str],
                                  period: Period,
                                  date_from: datetime.datetime,
                                  date_to: datetime.datetime,
                                  show_progress: bool,
                                  ):
        pass



from ziplime.constants.fundamental_data import FUNDAMENTAL_DATA_COLUMNS
from ziplime.domain.column_specification import ColumnSpecification

OHLCV_COLUMNS = [
    ColumnSpecification(name="close", write_type="uint32", original_type='float64', scale_factor=1000,
                        scaled_type="int"),
    ColumnSpecification(name="open", write_type="uint32", original_type='float64', scale_factor=1000,
                        scaled_type="int"),
    ColumnSpecification(name="high", write_type="uint32", original_type='float64', scale_factor=1000,
                        scaled_type="int"),
    ColumnSpecification(name="low", write_type="uint32", original_type='float64', scale_factor=1000,
                        scaled_type="int"),
    ColumnSpecification(name="volume", write_type="uint32", original_type='int', scale_factor=1,
                        scaled_type="int")
]

DEFAULT_COLUMNS = OHLCV_COLUMNS + FUNDAMENTAL_DATA_COLUMNS



import datetime
import os

from lime_trader.models.market import Period
from zipline.utils.paths import data_root

from ziplime.config.register_bundles import register_lime_symbol_list_equities_bundle
from ziplime.constants.bundles import DEFAULT_BUNDLE
from ziplime.constants.fundamental_data import FUNDAMENTAL_DATA_COLUMN_NAMES
from ziplime.data.bundles.core import BundleData
from ziplime.data.providers.fundamental_data.limex_hub_fundamental_data_provider import LimexHubFundamentalDataProvider
from ziplime.data.providers.historical_market_data.limex_hub_historical_market_data_provider import \
    LimexHubHistoricalMarketDataProvider
from ziplime.data.providers.live_market_data.lime_trader_sdk_live_market_data_provider import \
    LimeTraderSdkLiveMarketDataProvider


def register_default_bundles(calendar_name: str = "NYSE",
                             fundamental_data_list: set[str] = FUNDAMENTAL_DATA_COLUMN_NAMES):
    data_path = data_root()
    if not next(os.walk(data_path), None):
        lime_bundle_names = []
    else:
        lime_bundle_names = [x for x in [x for x in next(os.walk(data_path), [])][1] if x.startswith(DEFAULT_BUNDLE)]
    for bundle in lime_bundle_names:
        register_lime_symbol_list_equities_bundle(
            bundle_name=bundle,
            symbols=[],
            start_session=None,#datetime.datetime.now().replace(minute=0, hour=0, second=0, microsecond=0),
            end_session=None,#datetime.datetime.now().replace(minute=0, hour=0, second=0, microsecond=0),
            period=Period("day"),
            calendar_name=calendar_name,
            fundamental_data_list=fundamental_data_list
        )


def get_historical_market_data_provider(code: str):
    if code == "limex-hub":
        limex_hub_key = os.environ.get("LIMEX_API_KEY", None)
        if limex_hub_key is None:
            raise ValueError("Missing LIMEX_API_KEY environment variable.")
        return LimexHubHistoricalMarketDataProvider(limex_api_key=limex_hub_key)
    raise Exception("Unsupported historical market data provider!")

def get_live_market_data_provider(code: str):
    if code == "lime-trader-sdk":
        lime_trader_sdk_credentials = os.environ.get("LIME_SDK_CREDENTIALS_FILE", None)
        if lime_trader_sdk_credentials is None:
            raise ValueError("Missing LIME_SDK_CREDENTIALS_FILE environment variable.")
        return LimeTraderSdkLiveMarketDataProvider(lime_sdk_credentials_file=lime_trader_sdk_credentials)
    raise Exception("Unsupported live market data provider!")

def get_fundamental_data_provider(code: str):
    if code == "limex-hub":
        limex_hub_key = os.environ.get("LIMEX_API_KEY", None)
        if limex_hub_key is None:
            raise ValueError("Missing LIMEX_API_KEY environment variable.")
        return LimexHubFundamentalDataProvider(limex_api_key=limex_hub_key)
    raise Exception("Unsupported fundamental data provider!")


import datetime
import logging
import multiprocessing
import sys
from dataclasses import asdict
from queue import Queue

import limexhub
import numpy
import pandas as pd
from click import progressbar
from joblib import Parallel, delayed
from lime_trader import LimeClient
from lime_trader.models.market import Period

from ziplime.constants.fundamental_data import FundamentalData
from ziplime.data.abstract_fundamendal_data_provider import AbstractFundamentalDataProvider
from ziplime.domain.lime_quote import LimeQuote


class LimexHubFundamentalDataProvider(AbstractFundamentalDataProvider):
    def __init__(self, limex_api_key: str):
        self._limex_api_key = limex_api_key
        self._logger = logging.getLogger(__name__)
        self._limex_client = limexhub.RestAPI(token=limex_api_key)

    def get_fundamental_data(self,
                             symbols: list[str],
                             date_from: datetime.datetime,
                             date_to: datetime.datetime,
                             period: Period,
                             fundamental_data_list: set[str]
                             ):
        data = pd.DataFrame()
        for symbol in symbols:
            res = self._get_fundamental_data(symbol, date_from, date_to, period, fundamental_data_list)
            data = pd.concat([data, res], axis=0, ignore_index=False)
        data.set_index(["date", "symbol"], inplace=True)

        return data

    def _get_fundamental_data(self,
                              symbol: str,
                              date_from: datetime.datetime,
                              date_to: datetime.datetime,
                              period: Period,
                              fundamental_data_list: set[str]
                              ):

        # print("FUNDAMENTAL DATA")
        fundamental = self._limex_client.fundamental(
            symbol=symbol,
            from_date=date_from,
            to_date=date_to,
            fields=None
        )
        if fundamental.empty:
            logging.warning(f"No fundamental data for {symbol} for period {date_from}-{date_to}.")
        if not fundamental.empty:
            fundamental["date"] = pd.to_datetime(fundamental.date, utc=True)
        dr = pd.date_range(date_from, date_to, freq='D')
        fundamental_new = pd.DataFrame(columns=["date", "symbol"])
        fundamental_new.set_index(keys=["date", "symbol"], inplace=True, drop=True)
        use_datetime = True
        for fund_col in FundamentalData:

            # print(fund_col.value)

            col_name = fund_col.value
            ttm_col = f"{col_name}_ttm"
            value_col = f"{col_name}_value"
            add_value_col = f"{col_name}_add_value"
            columns = ["date", "symbol"]
            if ttm_col in fundamental_data_list:
                columns.append(ttm_col)
            if value_col in fundamental_data_list:
                columns.append(value_col)
            if add_value_col in fundamental_data_list:
                columns.append(add_value_col)

            res_df = pd.DataFrame(columns=columns)
            if len(columns) <= 2:
                continue
            if "field" in fundamental.columns:
                values_for_col = fundamental[fundamental.field == col_name]
                for index, row in values_for_col.iterrows():
                    if use_datetime:
                        dt = datetime.datetime.combine(row.date, time=datetime.time(), tzinfo=datetime.timezone.utc)
                    else:
                        dt = row.date

                    vals = [dt, symbol]
                    if ttm_col in columns:
                        vals.append(row.ttm)
                    if value_col in columns:
                        vals.append(row.value)
                    if add_value_col in columns:
                        vals.append(row.add_value)
                    vals_df = pd.DataFrame([
                        vals,
                    ], columns=res_df.columns)
                    res_df = (
                        res_df.copy()
                        if vals_df.empty
                        else vals_df
                        if res_df.empty
                        else pd.concat([res_df, vals_df], ignore_index=True)
                    )
                    # res_df = pd.concat([pd.DataFrame([
                    #     vals,
                    # ], columns=res_df.columns), res_df], ignore_index=True)
            else:
                vals = [date_from, symbol]
                if ttm_col in columns:
                    vals.append(numpy.nan)
                if value_col in columns:
                    vals.append(numpy.nan)
                if add_value_col in columns:
                    vals.append(numpy.nan)
                vals_df = pd.DataFrame([
                    vals,
                ], columns=res_df.columns)

                res_df = (
                    res_df.copy()
                    if vals_df.empty
                    else vals_df
                    if res_df.empty
                    else pd.concat([vals_df, res_df], ignore_index=True)
                )
            res_df.set_index("date", inplace=True, drop=True)
            res_df = res_df.reindex(dr, fill_value=None)
            res_df["symbol"] = symbol
            res_df.reset_index(inplace=True)
            res_df.rename(columns={"index": "date"}, inplace=True)
            res_df.set_index(["date", "symbol"], inplace=True)
            fundamental_new = pd.concat([res_df, fundamental_new], ignore_index=False, axis=1)

        fundamental_new.reset_index(inplace=True)
        filtered = fundamental_new[fundamental_new.symbol.notnull()]
        # filtered.set_index(["date", "symbol"], inplace=True)

        return filtered



#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging

from ziplime.data.abstract_live_market_data_provider import AbstractLiveMarketDataProvider
from ziplime.data.data_portal import DataPortal

log = logging.Logger('DataPortalLive')


class DataPortalLive(DataPortal):
    def __init__(self, broker, market_data_provider: AbstractLiveMarketDataProvider, *args, **kwargs):
        self.broker = broker
        self.market_data_provider = market_data_provider
        super(DataPortalLive, self).__init__(*args, **kwargs)

    def get_last_traded_dt(self, asset, dt, data_frequency):
        return self.broker.get_last_traded_dt(asset)

    def get_spot_value(self, assets, field, dt, data_frequency):
        return self.broker.get_spot_value(assets, field, dt, data_frequency)

    def get_history_window(self,
                           assets,
                           end_dt,
                           bar_count,
                           frequency,
                           field,
                           data_frequency,
                           ffill: bool = True):
        # This method is responsible for merging the ingested historical data
        # with the real-time collected data through the Broker.
        # DataPortal.get_history_window() is called with ffill=False to mark
        # the missing fields with NaNs. After merge on the historical and
        # real-time data the missing values (NaNs) are filled based on their
        # next available values in the requested time window.
        #
        # Warning: setting ffill=True in DataPortal.get_history_window() call
        # results a wrong behavior: The last available value reported by
        # get_spot_value() will be used to fill the missing data - which is
        # always representing the current spot price presented by Broker.

        historical_bars = super(DataPortalLive, self).get_history_window(
            assets=assets, end_dt=end_dt, bar_count=bar_count, frequency=frequency, field=field,
            data_frequency=data_frequency, ffill=False
        )

        realtime_bars = self.market_data_provider.fetch_live_data_table(
            symbols=assets, period=frequency, date_from=end_dt, date_to=end_dt,
            show_progress=False
        )

        # Broker.get_realtime_history() returns the asset as level 0 column,
        # open, high, low, close, volume returned as level 1 columns.
        # To filter for field the levels needs to be swapped
        realtime_bars = realtime_bars.swaplevel(0, 1, axis=1)

        ohlcv_field = 'close' if field == 'price' else field

        # TODO: end_dt is ignored when historical & realtime bars are merged.
        # Should not cause issues as end_dt is set to current time in live
        # trading, but would be more proper if merge would make use of it.
        combined_bars = historical_bars.combine_first(
            realtime_bars[ohlcv_field])

        if ffill and field == 'price':
            # Simple forward fill is not enough here as the last ingested
            # value might be outside of the requested time window. That case
            # the time series starts with NaN and forward filling won't help.
            # To provide values for such cases we backward fill.
            # Backward fill as a second operation will have no effect if the
            # forward-fill was successful.
            combined_bars.fillna(method='ffill', inplace=True)
            combined_bars.fillna(method='bfill', inplace=True)

        return combined_bars[-bar_count:]



# Copyright 2015 Quantopian, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import datetime
import warnings
from dataclasses import asdict
from functools import partial

import numpy
from exchange_calendars import ExchangeCalendar
from numpy import zeros
from zipline.data._equities import _read_bcolz_data, _compute_row_slices

from ziplime.constants.default_columns import DEFAULT_COLUMNS
from ziplime.data.abstract_data_bundle import AbstractDataBundle
from ziplime.domain.column_specification import ColumnSpecification

with warnings.catch_warnings():  # noqa
    warnings.filterwarnings("ignore", category=DeprecationWarning)
    from bcolz import carray, ctable
    import numpy as np

import logging

import pandas as pd

from zipline.data.bar_reader import NoDataAfterDate, NoDataBeforeDate, NoDataOnDate
from zipline.data.session_bars import CurrencyAwareSessionBarReader
from zipline.utils.calendar_utils import get_calendar
from zipline.utils.cli import maybe_show_progress
from zipline.utils.functional import apply
from zipline.utils.input_validation import expect_element
from zipline.utils.memoize import lazyval
from zipline.utils.numpy_utils import float64_dtype, iNaT, uint32_dtype

logger = logging.getLogger("UsEquityPricing")

# OHLC = frozenset(["open", "high", "low", "close"])
# US_EQUITY_PRICING_BCOLZ_COLUMNS = (
#     "open",
#     "high",
#     "low",
#     "close",
#     "volume",
#     "day",
#     "id",
# )
#
UINT32_MAX = np.iinfo(np.uint32).max


def check_uint32_safe(value, colname):
    if value >= UINT32_MAX:
        raise ValueError("Value %s from column '%s' is too large" % (value, colname))


@expect_element(invalid_data_behavior={"warn", "raise", "ignore"})
def winsorise_uint32(df, invalid_data_behavior, column, *columns):
    """Drops any record where a value would not fit into a uint32.

    Parameters
    ----------
    df : pd.DataFrame
        The dataframe to winsorise.
    invalid_data_behavior : {'warn', 'raise', 'ignore'}
        What to do when data is outside the bounds of a uint32.
    *columns : iterable[str]
        The names of the columns to check.

    Returns
    -------
    truncated : pd.DataFrame
        ``df`` with values that do not fit into a uint32 zeroed out.
    """
    columns = list((column,) + columns)
    mask = df[columns] > UINT32_MAX

    if invalid_data_behavior != "ignore":
        mask |= df[columns].isnull()
    else:
        # we are not going to generate a warning or error for this so just use
        # nan_to_num
        df[columns] = np.nan_to_num(df[columns])

    mv = mask.values
    if mv.any():
        if invalid_data_behavior == "raise":
            raise ValueError(
                "%d values out of bounds for uint32: %r"
                % (
                    mv.sum(),
                    df[mask.any(axis=1)],
                ),
            )
        if invalid_data_behavior == "warn":
            warnings.warn(
                "Ignoring %d values because they are out of bounds for"
                " uint32:\n %r"
                % (
                    mv.sum(),
                    df[mask.any(axis=1)],
                ),
                stacklevel=3,  # one extra frame for `expect_element`
            )

    df[mask] = 0
    return df


class BcolzDataBundle(AbstractDataBundle):
    """Class capable of writing daily OHLCV data to disk in a format that can
    be read efficiently by BcolzDailyOHLCVReader.

    Parameters
    ----------
    filename : str
        The location at which we should write our output.
    calendar : zipline.utils.calendar.trading_calendar
        Calendar to use to compute asset calendar offsets.
    start_session: pd.Timestamp
        Midnight UTC session label.
    end_session: pd.Timestamp
        Midnight UTC session label.

    See Also
    --------
    zipline.data.bcolz_daily_bars.BcolzDailyBarReader
    """

    def __init__(self, root_directory: str):
        super().__init__()
        self._root_directory = root_directory
        # Cache of fully read np.array for the carrays in the daily bar table.
        # raw_array does not use the same cache, but it could.
        # Need to test keeping the entire array in memory for the course of a
        # process first.
        self._spot_cols = {}

    @lazyval
    def _table(self) -> ctable:
        return ctable(rootdir=self._root_directory, mode="r")

    @lazyval
    def sessions(self):
        if "calendar" in self._table.attrs.attrs:
            # backwards compatibility with old formats, will remove
            return pd.DatetimeIndex(self._table.attrs["calendar"])
        else:
            cal = get_calendar(self._table.attrs["calendar_name"])
            start_session_ns = self._table.attrs["start_session_ns"]

            start_session = pd.Timestamp(start_session_ns)

            end_session_ns = self._table.attrs["end_session_ns"]
            end_session = pd.Timestamp(end_session_ns)

            sessions = cal.sessions_in_range(start_session, end_session)

            return sessions

    @lazyval
    def _first_rows(self):
        return {
            int(asset_id): start_index
            for asset_id, start_index in self._table.attrs["first_row"].items()
        }

    @lazyval
    def _last_rows(self):
        return {
            int(asset_id): end_index
            for asset_id, end_index in self._table.attrs["last_row"].items()
        }

    @lazyval
    def _calendar_offsets(self):
        return {
            int(id_): offset
            for id_, offset in self._table.attrs["calendar_offset"].items()
        }

    @lazyval
    def first_trading_day(self):
        try:
            return pd.Timestamp(self._table.attrs["first_trading_day"], unit="s")
        except KeyError:
            return None

    @lazyval
    def trading_calendar(self):
        if "calendar_name" in self._table.attrs.attrs:
            return get_calendar(self._table.attrs["calendar_name"])
        else:
            return None

    @property
    def last_available_dt(self):
        return self.sessions[-1]

    def _compute_slices(self, start_idx, end_idx, assets):
        """Compute the raw row indices to load for each asset on a query for the
        given dates after applying a shift.

        Parameters
        ----------
        start_idx : int
            Index of first date for which we want data.
        end_idx : int
            Index of last date for which we want data.
        assets : pandas.Int64Index
            Assets for which we want to compute row indices

        Returns
        -------
        A 3-tuple of (first_rows, last_rows, offsets):
        first_rows : np.array[intp]
            Array with length == len(assets) containing the index of the first
            row to load for each asset in `assets`.
        last_rows : np.array[intp]
            Array with length == len(assets) containing the index of the last
            row to load for each asset in `assets`.
        offset : np.array[intp]
            Array with length == (len(asset) containing the index in a buffer
            of length `dates` corresponding to the first row of each asset.

            The value of offset[i] will be 0 if asset[i] existed at the start
            of a query.  Otherwise, offset[i] will be equal to the number of
            entries in `dates` for which the asset did not yet exist.
        """
        # The core implementation of the logic here is implemented in Cython
        # for efficiency.
        return _compute_row_slices(
            self._first_rows,
            self._last_rows,
            self._calendar_offsets,
            start_idx,
            end_idx,
            assets,
        )

    @lazyval
    def _column_specification(self):
        cols = [
            ColumnSpecification(
                name=col["name"],
                original_type=col["original_type"],
                write_type=col["write_type"],
                scaled_type=col["scaled_type"],
                scale_factor=col["scale_factor"]
            )
            for col in self._table.attrs["column_specification"]
        ]
        return {
            col.name: col
            for col in cols
        }

    def load_raw_arrays(self, columns, start_date, end_date, assets):
        start_idx = self._load_raw_arrays_date_to_index(start_date)
        end_idx = self._load_raw_arrays_date_to_index(end_date)

        first_rows, last_rows, offsets = self._compute_slices(
            start_idx,
            end_idx,
            assets,
        )
        read_all = True
        return self._read_bcolz_data(
            self._table,
            (end_idx - start_idx + 1, len(assets)),
            list(columns),
            first_rows,
            last_rows,
            offsets,
            read_all,
        )
    def load_raw_arrays_full_range(self, columns, start_date, end_date, assets, frequency):
        # start_idx = self._load_raw_arrays_date_to_index(start_date)
        # end_idx = self._load_raw_arrays_date_to_index(end_date)
        dr = pd.date_range(pd.to_datetime(self._table.attrs["first_trading_day"], unit="s"), end_date, freq=frequency)
        start_idx = dr.get_loc(start_date)
        end_idx = dr.get_loc(end_date)
        first_rows, last_rows, offsets = self._compute_slices(
            start_idx,
            end_idx,
            assets,
        )
        read_all = True
        return self._read_bcolz_data(
            self._table,
            (end_idx - start_idx + 1, len(assets)),
            list(columns),
            first_rows,
            last_rows,
            offsets,
            read_all,
        )

    def _load_raw_arrays_date_to_index(self, date):
        try:
            # TODO get_loc is deprecated but get_indexer doesnt raise and error
            return self.sessions.get_loc(date)
        except KeyError as exc:
            raise NoDataOnDate(date) from exc

    def _spot_col(self, colname):
        """Get the colname from daily_bar_table and read all of it into memory,
        caching the result.

        Parameters
        ----------
        colname : string
            A name of a OHLCV carray in the daily_bar_table

        Returns
        -------
        array (uint32)
            Full read array of the carray in the daily_bar_table with the
            given colname.
        """
        try:
            col = self._spot_cols[colname]
        except KeyError:
            col = self._spot_cols[colname] = self._table[colname]
        return col

    def get_last_traded_dt(self, asset, day):
        volumes = self._spot_col("volume")

        search_day = day

        while True:
            try:
                ix = self.sid_day_index(asset, search_day)
            except NoDataBeforeDate:
                return pd.NaT
            except NoDataAfterDate:
                prev_day_ix = self.sessions.get_loc(search_day) - 1
                if prev_day_ix > -1:
                    search_day = self.sessions[prev_day_ix]
                continue
            except NoDataOnDate:
                return pd.NaT
            if volumes[ix] != 0:
                return search_day
            prev_day_ix = self.sessions.get_loc(search_day) - 1
            if prev_day_ix > -1:
                search_day = self.sessions[prev_day_ix]
            else:
                return pd.NaT

    def sid_day_index(self, sid, day):
        """

        Parameters
        ----------
        sid : int
            The asset identifier.
        day : datetime64-like
            Midnight of the day for which data is requested.

        Returns
        -------
        int
            Index into the data tape for the given sid and day.
            Raises a NoDataOnDate exception if the given day and sid is before
            or after the date range of the equity.
        """
        try:
            day_loc = self.sessions.get_loc(day)
        except Exception as exc:
            raise NoDataOnDate(
                "day={0} is outside of calendar={1}".format(day, self.sessions)
            ) from exc
        offset = day_loc - self._calendar_offsets[sid]
        if offset < 0:
            raise NoDataBeforeDate(
                "No data on or before day={0} for sid={1}".format(day, sid)
            )
        ix = self._first_rows[sid] + offset
        if ix > self._last_rows[sid]:
            raise NoDataAfterDate(
                "No data on or after day={0} for sid={1}".format(day, sid)
            )
        return ix

    def get_value(self, sid, dt, field):
        """

        Parameters
        ----------
        sid : int
            The asset identifier.
        day : datetime64-like
            Midnight of the day for which data is requested.
        colname : string
            The price field. e.g. ('open', 'high', 'low', 'close', 'volume')

        Returns
        -------
        float
            The spot price for colname of the given sid on the given day.
            Raises a NoDataOnDate exception if the given day and sid is before
            or after the date range of the equity.
            Returns -1 if the day is within the date range, but the price is
            0.
        """
        ix = self.sid_day_index(sid, dt)
        price = self._spot_col(field)[ix]
        if field != "volume":
            if price == 0:
                return np.nan
            else:
                return price * 0.001
        else:
            return price

    def currency_codes(self, sids):
        # XXX: This is pretty inefficient. This reader doesn't really support
        # country codes, so we always either return USD or None if we don't
        # know about the sid at all.
        first_rows = self._first_rows
        out = []
        for sid in sids:
            if sid in first_rows:
                out.append("USD")
            else:
                out.append(None)
        return np.array(out, dtype=object)

    def _read_bcolz_data(self, table, shape, columns, first_rows, last_rows, offsets, read_all):
        """
        Load raw bcolz data for the given columns and indices.

        Parameters
        ----------
        table : bcolz.ctable
            The table from which to read.
        shape : tuple (length 2)
            The shape of the expected output arrays.
        columns : list[str]
            List of column names to read.

        first_rows : ndarray[intp]
        last_rows : ndarray[intp]
        offsets : ndarray[intp
            Arrays in the format returned by _compute_row_slices.
        read_all : bool
            Whether to read_all sid data at once, or to read a silce from the
            carray for each sid.

        Returns
        -------
        results : list of ndarray
            A 2D array of shape `shape` for each column in `columns`.
        """
        results = []
        ndays = shape[0]
        nassets = shape[1]
        if not nassets == len(first_rows) == len(last_rows) == len(offsets):
            raise ValueError("Incompatible index arrays.")
        for column_name in columns:
            col = self._column_specification[column_name]
            outbuf = zeros(shape=shape, dtype=col.write_type)
            if read_all:
                raw_data = table[column_name][:]

                for asset in range(nassets):
                    first_row = first_rows[asset]
                    last_row = last_rows[asset]
                    offset = offsets[asset]
                    if first_row <= last_row:
                        outbuf[offset:offset + (last_row + 1 - first_row), asset] = \
                            raw_data[first_row:last_row + 1]
                    else:
                        continue
            else:
                carray = table[column_name]

                for asset in range(nassets):
                    first_row = first_rows[asset]
                    last_row = last_rows[asset]
                    offset = offsets[asset]
                    out_start = offset
                    out_end = (last_row - first_row) + offset + 1
                    if first_row <= last_row:
                        outbuf[offset:offset + (last_row + 1 - first_row), asset] = \
                            carray[first_row:last_row + 1]
                    else:
                        continue

            # if column_name in {'open', 'high', 'low', 'close'}:
            if col.write_type != col.original_type:
                where_nan = (outbuf == 0)
                outbuf_as_float = outbuf.astype(col.original_type) * 1 / col.scale_factor
                outbuf_as_float[where_nan] = numpy.nan
                results.append(outbuf_as_float)
            else:
                results.append(outbuf)
        return results

    ############## WRITE #########################
    @property
    def progress_bar_message(self):
        return "Merging daily equity files:"

    def progress_bar_item_show_func(self, value):
        return value if value is None else str(value[0])

    def write(
            self, data, calendar: ExchangeCalendar, start_session: pd.Timestamp, end_session: pd.Timestamp,
            cols: list[ColumnSpecification],
            validate_sessions: bool,
            assets=None, show_progress=False, invalid_data_behavior="warn",
            **kwargs
    ):
        """

        Parameters
        ----------
        data : iterable[tuple[int, pandas.DataFrame or bcolz.ctable]]
            The data chunks to write. Each chunk should be a tuple of sid
            and the data for that asset.
        assets : set[int], optional
            The assets that should be in ``data``. If this is provided
            we will check ``data`` against the assets and provide better
            progress information.
        show_progress : bool, optional
            Whether or not to show a progress bar while writing.
        invalid_data_behavior : {'warn', 'raise', 'ignore'}, optional
            What to do when data is encountered that is outside the range of
            a uint32.

        Returns
        -------
        table : bcolz.ctable
            The newly-written table.
        """
        ctx = maybe_show_progress(
            ((sid, self.to_ctable(df, invalid_data_behavior, cols)) for sid, df in data),
            show_progress=show_progress,
            item_show_func=self.progress_bar_item_show_func,
            label=self.progress_bar_message,
            length=len(assets) if assets is not None else None,
        )
        with ctx as it:
            return self._write_internal(iterator=it, assets=assets, calendar=calendar,
                                        start_session=start_session, end_session=end_session,
                                        cols=cols, show_progress=show_progress,
                                        validate_sessions=validate_sessions
                                        )

    def _write_internal(self, iterator, assets,
                        calendar: ExchangeCalendar, start_session: pd.Timestamp, end_session: pd.Timestamp,
                        show_progress,
                        cols: list[ColumnSpecification],
                        validate_sessions: bool
                        ):
        """Internal implementation of write.

        `iterator` should be an iterator yielding pairs of (asset, ctable).
        """
        total_rows = 0
        first_row = {}
        last_row = {}
        calendar_offset = {}

        # Maps column name -> output carray.
        columns = {
            k.name: carray(np.array([], dtype=np.dtype(k.write_type)))
            for k in cols  # US_EQUITY_PRICING_BCOLZ_COLUMNS
        }

        earliest_date = None
        sessions = calendar.sessions_in_range(
            start=start_session, end=end_session
        )

        if assets is not None:

            @apply
            def iterator(iterator=iterator, assets=set(assets)):
                for asset_id, table in iterator:
                    if asset_id not in assets:
                        raise ValueError("unknown asset id %r" % asset_id)
                    yield asset_id, table

        for asset_id, table in iterator:
            nrows = len(table)
            for column_name in columns:
                if column_name == "id":
                    # We know what the content of this column is, so don't
                    # bother reading it.
                    columns["id"].append(
                        np.full((nrows,), asset_id, dtype="uint32"),
                    )
                    continue

                columns[column_name].append(table[column_name])

            if earliest_date is None:
                earliest_date = table["day"][0]
            else:
                earliest_date = min(earliest_date, table["day"][0])

            # Bcolz doesn't support ints as keys in `attrs`, so convert
            # assets to strings for use as attr keys.
            asset_key = str(asset_id)

            # Calculate the index into the array of the first and last row
            # for this asset. This allows us to efficiently load single
            # assets when querying the data back out of the table.
            first_row[asset_key] = total_rows
            last_row[asset_key] = total_rows + nrows - 1
            total_rows += nrows

            asset_first_day = pd.Timestamp(table["day"][0], unit="s").normalize()
            asset_last_day = pd.Timestamp(table["day"][-1], unit="s").normalize()

            asset_sessions = sessions[
                sessions.slice_indexer(asset_first_day, asset_last_day)
            ]
            if validate_sessions and len(table) != len(asset_sessions):
                missing_sessions = asset_sessions.difference(
                    pd.to_datetime(np.array(table["day"]), unit="s")
                ).tolist()

                extra_sessions = (
                    pd.to_datetime(np.array(table["day"]), unit="s")
                    .difference(asset_sessions)
                    .tolist()
                )
                raise AssertionError(
                    f"Got {len(table)} rows for daily bars table with "
                    f"first day={asset_first_day.date()}, last "
                    f"day={asset_last_day.date()}, expected {len(asset_sessions)} rows.\n"
                    f"Missing sessions: {missing_sessions}\nExtra sessions: {extra_sessions}"
                )

            # Calculate the number of trading days between the first date
            # in the stored data and the first date of **this** asset. This
            # offset used for output alignment by the reader.
            calendar_offset[asset_key] = sessions.get_loc(asset_first_day)

        # This writes the table to disk.
        full_table = ctable(
            columns=[columns[col.name] for col in cols],
            names=[col.name for col in cols],
            rootdir=self._root_directory,
            mode="w",
        )

        full_table.attrs["first_trading_day"] = (
            earliest_date if earliest_date is not None else iNaT
        )

        full_table.attrs["first_row"] = first_row
        full_table.attrs["last_row"] = last_row
        full_table.attrs["calendar_offset"] = calendar_offset
        full_table.attrs["calendar_name"] = calendar.name
        full_table.attrs["start_session_ns"] = start_session.value
        full_table.attrs["end_session_ns"] = end_session.value
        full_table.attrs["column_specification"] = [asdict(col) for col in cols]
        full_table.flush()
        return full_table

    @expect_element(invalid_data_behavior={"warn", "raise", "ignore"})
    def to_ctable(self, raw_data, invalid_data_behavior, cols):
        if isinstance(raw_data, ctable):
            # we already have a ctable so do nothing
            return raw_data
        data = raw_data.copy()
        for col in cols:
            if col.write_type != col.original_type:
                data[col.name] = (raw_data[col.name] * col.scale_factor).round().astype(col.scaled_type).astype(
                    col.write_type)
        dates = data.index.values.astype("datetime64[s]")
        check_uint32_safe(dates.max().view(np.int64), "day")
        data["day"] = dates.astype("uint32")
        return ctable.fromdataframe(data)



import datetime
import logging
import multiprocessing
import sys
from dataclasses import asdict
from queue import Queue

import limexhub
import numpy
import pandas as pd
from click import progressbar
from joblib import Parallel, delayed
from lime_trader import LimeClient
from lime_trader.models.market import Period

from ziplime.constants.fundamental_data import FundamentalData
from ziplime.domain.lime_quote import LimeQuote


class LimeDataProvider:
    def __init__(self, limex_api_key: str, lime_sdk_credentials_file: str):
        self._limex_api_key = limex_api_key
        self._lime_sdk_credentials_file = lime_sdk_credentials_file

        self._logger = logging.getLogger(__name__)
        self._limex_client = limexhub.RestAPI(token=limex_api_key)
        self._lime_sdk_client = LimeClient.from_file(lime_sdk_credentials_file, logger=self._logger)

    def get_fundamental_data(self, limex_client: limexhub.RestAPI,
                             symbol: str,
                             date_from: datetime.datetime,
                             date_to: datetime.datetime,
                             period: Period,
                             fundamental_data_list: set[str]
                             ):
        # print("FUNDAMENTAL DATA")
        fundamental = limex_client.fundamental(
            symbol=symbol,
            from_date=date_from,
            to_date=date_to,
            fields=None
        )

        dr = pd.date_range(date_from, date_to, freq='D')
        fundamental_new = pd.DataFrame(columns=["date", "symbol"])
        fundamental_new.set_index(keys=["date", "symbol"], inplace=True, drop=True)
        use_datetime = True
        for fund_col in FundamentalData:

            # print(fund_col.value)

            col_name = fund_col.value
            ttm_col = f"{col_name}_ttm"
            value_col = f"{col_name}_value"
            add_value_col = f"{col_name}_add_value"
            columns = ["date", "symbol"]
            if ttm_col in fundamental_data_list:
                columns.append(ttm_col)
            if value_col in fundamental_data_list:
                columns.append(value_col)
            if add_value_col in fundamental_data_list:
                columns.append(add_value_col)

            res_df = pd.DataFrame(columns=columns)

            if "field" in fundamental.columns:
                values_for_col = fundamental[fundamental.field == col_name]
                for index, row in values_for_col.iterrows():
                    if use_datetime:
                        dt = datetime.datetime.combine(row.date, time=datetime.time(), tzinfo=datetime.timezone.utc)
                    else:
                        dt = row.date

                    vals = [dt, symbol]
                    if ttm_col in columns:
                        vals.append(row.ttm)
                    if value_col in columns:
                        vals.append(row.value)
                    if add_value_col in columns:
                        vals.append(row.add_value)

                    res_df = pd.concat([pd.DataFrame([
                        vals,
                    ], columns=res_df.columns), res_df], ignore_index=True)
            else:
                vals = [date_from, symbol]
                if ttm_col in columns:
                    vals.append(numpy.nan)
                if value_col in columns:
                    vals.append(numpy.nan)
                if add_value_col in columns:
                    vals.append(numpy.nan)

                res_df = pd.concat([pd.DataFrame([
                    vals,
                ], columns=res_df.columns), res_df], ignore_index=True)
            res_df.set_index("date", inplace=True, drop=True)
            # if fund_col.value == "dividend_yield":
            #     print("A")
            res_df = res_df.reindex(dr, fill_value=None)  # .ffill().bfill()
            res_df["symbol"] = symbol
            res_df.reset_index(inplace=True)
            res_df.rename(columns={"index": "date"}, inplace=True)
            res_df.set_index(["date", "symbol"], inplace=True)
            fundamental_new = pd.concat([res_df, fundamental_new], ignore_index=False, axis=1)

        fundamental_new.reset_index(inplace=True)
        filtered = fundamental_new[fundamental_new.symbol.notnull()]
        filtered.set_index(["date"], inplace=True)

        return filtered

    def fetch_historical_data_table(self, symbols: list[str],
                                    period: Period,
                                    date_from: datetime.datetime,
                                    date_to: datetime.datetime,
                                    show_progress: bool,
                                    fundamental_data_list: set[str]
                                    ):

        def fetch_historical(limex_api_key: str, symbol: str):
            limex_client = limexhub.RestAPI(token=limex_api_key)
            timeframe = 3
            if period == Period.MINUTE:
                timeframe = 1
            elif period == Period.HOUR:
                timeframe = 2
            elif period == Period.DAY:
                timeframe = 3
            elif period == Period.WEEK:
                timeframe = 4
            elif period == Period.MONTH:
                timeframe = 5
            elif period == Period.QUARTER:
                timeframe = 6
            df = limex_client.candles(symbol=symbol,
                                      from_date=date_from,
                                      to_date=date_to,
                                      timeframe=timeframe)

            fundamental = self.get_fundamental_data(limex_client, symbol, date_from, date_to, period=period,
                                                    fundamental_data_list=fundamental_data_list)
            if len(df) > 0:
                df = df.reset_index()
                df = df.rename(
                    columns={"o": "open", "h": "high", "l": "low", "c": "close", "v": "volume", "Date": "date"})
                df = df.set_index('date', drop=False)
                df.index = pd.to_datetime(df.index, utc=True)
                df['dividend'] = 0
                df['split'] = 0
                final_df = pd.concat([df, fundamental], ignore_index=False, axis=1)
                final_df = final_df[final_df.date.notnull()]
                final_df["symbol"] = symbol
                return final_df
            return df

        total_days = (date_to - date_from).days
        final = pd.DataFrame()

        if show_progress:
            with progressbar(length=len(symbols) * total_days, label="Downloading historical data from LimexHub",
                             file=sys.stdout) as pbar:
                res = Parallel(n_jobs=multiprocessing.cpu_count() * 2, prefer="threads",
                               return_as="generator_unordered")(
                    delayed(fetch_historical)(self._limex_api_key, symbol) for symbol in symbols)
                for item in res:
                    pbar.update(total_days)
                    final = pd.concat([final, item])
        else:
            res = Parallel(n_jobs=multiprocessing.cpu_count() * 2, prefer="threads", return_as="generator_unordered")(
                delayed(fetch_historical)(self._limex_api_key, symbol) for symbol in symbols)
            for item in res:
                final = pd.concat([final, item])
        final = final.sort_index()
        return final

    def fetch_data_table(
            self,
            symbols: list[str],
            period: Period,
            date_from: datetime.datetime,
            date_to: datetime.datetime,
            show_progress: bool,
            fundamental_data_list: list[str]):
        historical_data = self.fetch_historical_data_table(symbols=symbols, period=period, date_from=date_from,
                                                           date_to=date_to,
                                                           show_progress=show_progress,
                                                           fundamental_data_list=fundamental_data_list)

        yield historical_data

        # live_data_start_date = max(historical_data.index)[0].to_pydatetime().replace(
        #     tzinfo=datetime.timezone.utc) if len(historical_data) > 0 else date_from

        # for quotes in self.fetch_live_data_table(symbols=symbols, period=period, date_from=live_data_start_date,
        #                                          date_to=date_to,
        #                                          show_progress=show_progress):
        #     yield quotes

    def fetch_live_data_table(
            self,
            symbols: list[str],
            period: Period,
            date_from: datetime.datetime,
            date_to: datetime.datetime,
            show_progress: bool):

        live_data_queue = Queue()

        def fetch_live(lime_trader_sdk_credentials_file: str, symbol: str):
            lime_client = LimeClient.from_file(lime_trader_sdk_credentials_file, logger=self._logger)
            try:
                quotes = lime_client.market.get_quotes_history(
                    symbol=symbol, period=period, from_date=date_from,
                    to_date=date_to
                )
                df = LimeDataProvider.load_data_table(
                    quotes=[LimeQuote(symbol=symbol, quote_history=quote) for quote in quotes],
                    show_progress=show_progress
                )
            except Exception as e:
                self._logger.error("Error fetching data using lime trader sdk")
                df = pd.DataFrame()
            live_data_queue.put(df)

        res = Parallel(n_jobs=multiprocessing.cpu_count() * 2, prefer="threads", return_as="generator_unordered", )(
            delayed(fetch_live)(self._lime_sdk_credentials_file, symbol) for symbol in symbols)

        if show_progress:
            self._logger.info("Downloading live Lime Trader SDK metadata.")
        processed_symbols = 0
        while True:
            if processed_symbols == len(symbols):
                break
            item = live_data_queue.get()
            yield item
            processed_symbols += 1

    @staticmethod
    def load_data_table(quotes: list[LimeQuote], show_progress: bool = False):
        if not quotes:
            return pd.DataFrame()
        data_table = pd.DataFrame(
            [dict(**asdict(quote_hist.quote_history), symbol=quote_hist.symbol) for quote_hist in quotes], )

        data_table.rename(
            columns={
                "timestamp": "date",
            },
            inplace=True,
            copy=False,
        )
        # data_table = data_table.reset_index()
        data_table = data_table.set_index('date', drop=False)
        return data_table



# Copyright 2015 Quantopian, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import datetime
import warnings
from functools import partial

from exchange_calendars import ExchangeCalendar
from zipline.data._equities import _read_bcolz_data, _compute_row_slices

from ziplime.constants.default_columns import DEFAULT_COLUMNS
from ziplime.domain.column_specification import ColumnSpecification

with warnings.catch_warnings():  # noqa
    warnings.filterwarnings("ignore", category=DeprecationWarning)
    from bcolz import carray, ctable
    import numpy as np

import logging

import pandas as pd

from zipline.utils.input_validation import expect_element

logger = logging.getLogger("UsEquityPricing")

OHLC = frozenset(["open", "high", "low", "close"])
# US_EQUITY_PRICING_BCOLZ_COLUMNS = (
#     "open",
#     "high",
#     "low",
#     "close",
#     "volume",
#     "day",
#     "id",
# )
#
UINT32_MAX = np.iinfo(np.uint32).max


def check_uint32_safe(value, colname):
    if value >= UINT32_MAX:
        raise ValueError("Value %s from column '%s' is too large" % (value, colname))


@expect_element(invalid_data_behavior={"warn", "raise", "ignore"})
def winsorise_uint32(df, invalid_data_behavior, column, *columns):
    """Drops any record where a value would not fit into a uint32.

    Parameters
    ----------
    df : pd.DataFrame
        The dataframe to winsorise.
    invalid_data_behavior : {'warn', 'raise', 'ignore'}
        What to do when data is outside the bounds of a uint32.
    *columns : iterable[str]
        The names of the columns to check.

    Returns
    -------
    truncated : pd.DataFrame
        ``df`` with values that do not fit into a uint32 zeroed out.
    """
    columns = list((column,) + columns)
    mask = df[columns] > UINT32_MAX

    if invalid_data_behavior != "ignore":
        mask |= df[columns].isnull()
    else:
        # we are not going to generate a warning or error for this so just use
        # nan_to_num
        df[columns] = np.nan_to_num(df[columns])

    mv = mask.values
    if mv.any():
        if invalid_data_behavior == "raise":
            raise ValueError(
                "%d values out of bounds for uint32: %r"
                % (
                    mv.sum(),
                    df[mask.any(axis=1)],
                ),
            )
        if invalid_data_behavior == "warn":
            warnings.warn(
                "Ignoring %d values because they are out of bounds for"
                " uint32:\n %r"
                % (
                    mv.sum(),
                    df[mask.any(axis=1)],
                ),
                stacklevel=3,  # one extra frame for `expect_element`
            )

    df[mask] = 0
    return df


class ZiplimeBcolzDailyBarWriter:
    """Class capable of writing daily OHLCV data to disk in a format that can
    be read efficiently by BcolzDailyOHLCVReader.

    Parameters
    ----------
    filename : str
        The location at which we should write our output.
    calendar : zipline.utils.calendar.trading_calendar
        Calendar to use to compute asset calendar offsets.
    start_session: pd.Timestamp
        Midnight UTC session label.
    end_session: pd.Timestamp
        Midnight UTC session label.

    See Also
    --------
    zipline.data.bcolz_daily_bars.BcolzDailyBarReader
    """

    # _csv_dtypes = {
    #     "open": float64_dtype,
    #     "high": float64_dtype,
    #     "low": float64_dtype,
    #     "close": float64_dtype,
    #     "volume": float64_dtype,
    # }

    def __init__(self, filename: str, calendar: ExchangeCalendar, start_session: pd.Timestamp,
                 end_session: pd.Timestamp, cols: list[ColumnSpecification] = DEFAULT_COLUMNS):

        self.cols = cols
        self._filename = filename
        start_session = start_session.tz_localize(None)
        end_session = end_session.tz_localize(None)

        if start_session != end_session:
            if not calendar.is_session(start_session):
                raise ValueError("Start session %s is invalid!" % start_session)
            if not calendar.is_session(end_session):
                raise ValueError("End session %s is invalid!" % end_session)

        self._start_session = start_session
        self._end_session = end_session

        self._calendar = calendar



import datetime
import logging
import os
from functools import partial

import pandas as pd
from zipline.__main__ import ipython_only

from zipline.utils.calendar_utils import get_calendar
from zipline.utils.cli import Date

from ziplime.constants.default_columns import DEFAULT_COLUMNS, OHLCV_COLUMNS
from ziplime.constants.fundamental_data import FUNDAMENTAL_DATA_COLUMNS
from ziplime.data.storages.bcolz_data_bundle import BcolzDataBundle
from ziplime.utils.run_algo import _run, BenchmarkSpec

import click
import zipline
from ziplime.data import bundles as bundles_module

from click import DateTime
from lime_trader.models.market import Period
from zipline.utils.cli import Timestamp
from zipline.utils.paths import data_root

from ziplime.config.register_bundles import register_lime_symbol_list_equities_bundle

from zipline import __main__ as zipline__main__
from zipline.utils.run_algo import load_extensions
from zipline.extensions import create_args

from ziplime.data.bcolz_daily_bars import ZiplimeBcolzDailyBarWriter
from ziplime.data.bcolz_minute_bars import ZiplimeBcolzMinuteBarWriter
from ziplime.utils.bundle_utils import register_default_bundles, get_historical_market_data_provider, \
    get_fundamental_data_provider, get_live_market_data_provider

DEFAULT_BUNDLE = "lime"


def validate_date_range(date_min: datetime.datetime, date_max: datetime.datetime):
    def _validate_date_range(ctx, param, value):
        if not date_min < value.replace(tzinfo=datetime.timezone.utc) < date_max:
            raise click.BadParameter(f"Must be between {date_min} and {date_max}")
        return value

    return _validate_date_range


@click.group()
@click.option(
    "-e",
    "--extension",
    multiple=True,
    help="File or module path to a zipline extension to load.",
)
@click.option(
    "--strict-extensions/--non-strict-extensions",
    is_flag=True,
    help="If --strict-extensions is passed then zipline will not "
         "run if it cannot load all of the specified extensions. "
         "If this is not passed or --non-strict-extensions is passed "
         "then the failure will be logged but execution will continue.",
)
@click.option(
    "--default-extension/--no-default-extension",
    is_flag=True,
    default=True,
    help="Don't load the default zipline extension.py file in $ZIPLINE_HOME.",
)
@click.option(
    "-x",
    multiple=True,
    help="Any custom command line arguments to define, in key=value form.",
)
@click.pass_context
def main(ctx, extension, strict_extensions, default_extension, x):
    """Top level ziplime entry point."""
    # install a logging handler before performing any other operations

    logging.basicConfig(
        format="[%(asctime)s-%(levelname)s][%(name)s]\n %(message)s",
        level=logging.INFO,
        datefmt="%Y-%m-%dT%H:%M:%S%z",
    )

    register_default_bundles()

    create_args(x, zipline.extension_args)
    load_extensions(
        default_extension,
        extension,
        strict_extensions,
        os.environ,
    )


@main.command(context_settings=dict(
    # ignore_unknown_options=True,
    # allow_extra_args=True,
))
@click.option(
    "-b",
    "--bundle",
    default=DEFAULT_BUNDLE,
    metavar="BUNDLE-NAME",
    show_default=True,
    help="The data bundle to ingest.",
)
@click.option(
    "-c",
    "--calendar",
    default="NYSE",
    show_default=True,
    help="Default calendar to use.",
)
@click.option('--new-bundle-name', default=None)
@click.option(
    "--start-date",
    type=DateTime(),
    callback=validate_date_range(date_min=datetime.datetime(year=1800,
                                                            month=1,
                                                            day=1,
                                                            tzinfo=datetime.timezone.utc
                                                            ),
                                 date_max=datetime.datetime.now(tz=datetime.timezone.utc),
                                 )
)
@click.option(
    "--end-date",
    type=DateTime(),
    callback=validate_date_range(date_min=datetime.datetime(year=1800,
                                                            month=1,
                                                            day=1,
                                                            tzinfo=datetime.timezone.utc
                                                            ),
                                 date_max=datetime.datetime.now(tz=datetime.timezone.utc),
                                 )
)
@click.option(
    "--period",
    default="day",
    type=click.Choice(['minute', 'hour', 'day', 'week', 'month', 'quarter'], case_sensitive=False),
)
@click.option("-s", '--symbols')
@click.option('--fundamental-data')
@click.option(
    "--assets-version",
    type=int,
    multiple=True,
    help="Version of the assets db to which to downgrade.",
)
@click.option(
    "--show-progress/--no-show-progress",
    default=True,
    help="Print progress information to the terminal.",
)
@click.option(
    "--historical-market-data-provider",
    default="limex-hub",
    help="Market data provider for historical data",
    show_default=True,
)
@click.option(
    "--fundamental-data-provider",
    default="limex-hub",
    help="Fundamental data provider",
    show_default=True,
)
@click.pass_context
def ingest(ctx, bundle, new_bundle_name, start_date, end_date, period, symbols, fundamental_data, show_progress,
           assets_version, calendar,
           historical_market_data_provider,
           fundamental_data_provider
           ):
    """Top level ziplime entry point."""
    symbols_parsed = symbols.split(',') if symbols else None
    fundamental_data_list = fundamental_data.split(",") if fundamental_data else None
    if new_bundle_name:
        bundle_name = f"{DEFAULT_BUNDLE}-{new_bundle_name}"
        ctx.args = ['-b', new_bundle_name] + ctx.args
    else:
        bundle_name = bundle
        ctx.args = ['-b', bundle] + ctx.args
    # install a logging handler before performing any other operations
    register_lime_symbol_list_equities_bundle(
        bundle_name=bundle_name,
        symbols=symbols_parsed,
        start_session=start_date,
        end_session=end_date,
        period=Period(period),
        calendar_name=calendar,
        fundamental_data_list=fundamental_data_list if fundamental_data_list is not None else [
            col.name for col in
            FUNDAMENTAL_DATA_COLUMNS
        ]
    )

    new_params = dict(**ctx.params)

    # clean up lime only params and set new bundle name
    new_params["bundle"] = bundle_name

    fundamental_data_cols = ([
                                 col
                                 for col in FUNDAMENTAL_DATA_COLUMNS
                                 if col.name in set(fundamental_data_list)
                             ]
                             if fundamental_data_list is not None
                             else DEFAULT_COLUMNS)
    bundles_module.ingest(
        name=bundle_name,
        environ=os.environ,
        timestamp=pd.Timestamp.utcnow(),
        assets_version=assets_version,
        show_progress=show_progress,
        historical_market_data_provider=get_historical_market_data_provider(code=historical_market_data_provider),
        fundamental_data_provider=get_fundamental_data_provider(code=fundamental_data_provider),
        minute_bar_writer_class=ZiplimeBcolzMinuteBarWriter,
        daily_bar_writer_class=BcolzDataBundle,
        fundamental_data_writer_class=BcolzDataBundle,
        market_data_fields=OHLCV_COLUMNS,
        fundamental_data_fields=fundamental_data_cols,
        minute_bar_writer_cols=fundamental_data_cols,
    )


@main.command(context_settings=dict(
    ignore_unknown_options=True,
    allow_extra_args=True,
))
@click.option(
    "-b",
    "--bundle",
    default=DEFAULT_BUNDLE,
    metavar="BUNDLE-NAME",
    show_default=True,
    help="The data bundle to clean.",
)
@click.option(
    "-e",
    "--before",
    type=Timestamp(),
    help="Clear all data before TIMESTAMP."
         " This may not be passed with -k / --keep-last",
)
@click.option(
    "-a",
    "--after",
    type=Timestamp(),
    help="Clear all data after TIMESTAMP"
         " This may not be passed with -k / --keep-last",
)
@click.option(
    "-k",
    "--keep-last",
    type=int,
    metavar="N",
    help="Clear all but the last N downloads."
         " This may not be passed with -e / --before or -a / --after",
)
@click.pass_context
def clean(ctx, bundle, before, after, keep_last):
    """Top level ziplime entry point."""

    func = getattr(zipline__main__, "clean")
    ctx.forward(func)


@main.command(context_settings=dict(
    ignore_unknown_options=True,
    allow_extra_args=True,
))
@click.pass_context
def bundles(ctx):
    """Top level ziplime entry point."""
    func = getattr(zipline__main__, "bundles")
    ctx.forward(func)


@main.command()
@click.option(
    "-f",
    "--algofile",
    default=None,
    type=click.File("r"),
    help="The file that contains the algorithm to run.",
)
@click.option(
    "-t",
    "--algotext",
    help="The algorithm script to run.",
)
@click.option(
    "-D",
    "--define",
    multiple=True,
    help="Define a name to be bound in the namespace before executing"
         " the algotext. For example '-Dname=value'. The value may be any "
         "python expression. These are evaluated in order so they may refer "
         "to previously defined names.",
)
@click.option(
    "--data-frequency",
    type=click.Choice({"daily", "minute"}),
    default="daily",
    show_default=True,
    help="The data frequency of the simulation.",
)
@click.option(
    "--capital-base",
    type=float,
    default=10e6,
    show_default=True,
    help="The starting capital for the simulation.",
)
@click.option(
    "-b",
    "--bundle",
    default=DEFAULT_BUNDLE,
    metavar="BUNDLE-NAME",
    show_default=True,
    help="The data bundle to use for the simulation.",
)
@click.option(
    "--bundle-timestamp",
    type=Timestamp(),
    default=pd.Timestamp.utcnow(),
    show_default=False,
    help="The date to lookup data on or before.\n" "[default: <current-time>]",
)
@click.option(
    "-bf",
    "--benchmark-file",
    default=None,
    type=click.Path(exists=True, dir_okay=False, readable=True, path_type=str),
    help="The csv file that contains the benchmark returns",
)
@click.option(
    "--benchmark-symbol",
    default=None,
    type=click.STRING,
    help="The symbol of the instrument to be used as a benchmark "
         "(should exist in the ingested bundle)",
)
@click.option(
    "--benchmark-sid",
    default=None,
    type=int,
    help="The sid of the instrument to be used as a benchmark "
         "(should exist in the ingested bundle)",
)
@click.option(
    "--no-benchmark",
    is_flag=True,
    default=False,
    help="If passed, use a benchmark of zero returns.",
)
@click.option(
    "-s",
    "--start",
    type=Date(as_timestamp=True),
    help="The start date of the simulation.",
)
@click.option(
    "-e",
    "--end",
    type=Date(as_timestamp=True),
    help="The end date of the simulation.",
)
@click.option(
    "-o",
    "--output",
    default="-",
    metavar="FILENAME",
    show_default=True,
    help="The location to write the perf data. If this is '-' the perf will"
         " be written to stdout.",
)
@click.option(
    "--trading-calendar",
    metavar="TRADING-CALENDAR",
    default="XNYS",
    help="The calendar you want to use e.g. XLON. XNYS is the default.",
)
@click.option(
    "--print-algo/--no-print-algo",
    is_flag=True,
    default=False,
    help="Print the algorithm to stdout.",
)
@click.option(
    "--metrics-set",
    default="default",
    help="The metrics set to use. New metrics sets may be registered in your"
         " extension.py.",
)
@click.option(
    "--blotter",
    default="default",
    help="The blotter to use.",
    show_default=True,
)
@click.option(
    "--broker",
    default=None,
    help="The broker to use for live trading.",
    show_default=True,
)
@click.option(
    "--live-market-data-provider",
    default="lime-trader-sdk",
    help="Market data provider for live trading",
    show_default=True,
)
@ipython_only(
    click.option(
        "--local-namespace/--no-local-namespace",
        is_flag=True,
        default=None,
        help="Should the algorithm methods be " "resolved in the local namespace.",
    )
)
@click.pass_context
def run(
        ctx,
        algofile,
        algotext,
        define,
        data_frequency,
        capital_base,
        bundle,
        bundle_timestamp,
        benchmark_file,
        benchmark_symbol,
        benchmark_sid,
        no_benchmark,
        start,
        end,
        output,
        trading_calendar,
        print_algo,
        metrics_set,
        local_namespace,
        blotter,
        broker: str | None,
        live_market_data_provider: str,
):
    """Run a backtest for the given algorithm."""
    # check that the start and end dates are passed correctly
    if start is None and end is None:
        # check both at the same time to avoid the case where a user
        # does not pass either of these and then passes the first only
        # to be told they need to pass the second argument also
        ctx.fail(
            "must specify dates with '-s' / '--start' and '-e' / '--end'",
        )
    if start is None:
        ctx.fail("must specify a start date with '-s' / '--start'")
    if end is None:
        ctx.fail("must specify an end date with '-e' / '--end'")

    if (algotext is not None) == (algofile is not None):
        ctx.fail(
            "must specify exactly one of '-f' / "
            "'--algofile' or"
            " '-t' / '--algotext'",
        )

    trading_calendar = get_calendar(trading_calendar)

    benchmark_spec = BenchmarkSpec.from_cli_params(
        no_benchmark=no_benchmark,
        benchmark_sid=benchmark_sid,
        benchmark_symbol=benchmark_symbol,
        benchmark_file=benchmark_file,
    )

    return _run(
        initialize=None,
        handle_data=None,
        before_trading_start=None,
        analyze=None,
        algofile=algofile,
        algotext=algotext,
        defines=define,
        data_frequency=data_frequency,
        capital_base=capital_base,
        bundle=bundle,
        bundle_timestamp=bundle_timestamp,
        start=start,
        end=end,
        output=output,
        trading_calendar=trading_calendar,
        print_algo=print_algo,
        metrics_set=metrics_set,
        local_namespace=local_namespace,
        environ=os.environ,
        blotter=blotter,
        benchmark_spec=benchmark_spec,
        custom_loader=None,
        broker=broker,
        market_data_provider=get_live_market_data_provider(live_market_data_provider)
    )


if __name__ == "__main__":
    main()



import datetime
from os import environ, _Environ
import logging
import pandas as pd

from exchange_calendars import ExchangeCalendar
from lime_trader.models.market import Period
from zipline.assets import AssetDBWriter
from zipline.data.adjustments import SQLiteAdjustmentWriter
from zipline.data.bcolz_daily_bars import BcolzDailyBarWriter
from zipline.data.bcolz_minute_bars import BcolzMinuteBarWriter
from zipline.utils.cache import dataframe_cache
from zipline.utils.calendar_utils import register_calendar_alias

from ziplime.constants.default_columns import OHLCV_COLUMNS
from ziplime.data.abstract_fundamendal_data_provider import AbstractFundamentalDataProvider
from ziplime.data.abstract_historical_market_data_provider import AbstractHistoricalMarketDataProvider
from ziplime.data.bundles import core as bundles, register
import numpy as np

from ziplime.data.lime_data_provider import LimeDataProvider
from ziplime.data.storages.bcolz_data_bundle import BcolzDataBundle
from ziplime.domain.column_specification import ColumnSpecification
from ziplime.utils.calendar_utils import normalize_daily_start_end_session

logger = logging.getLogger(__name__)


def gen_asset_metadata(data: pd.DataFrame, show_progress: bool):
    if show_progress:
        logger.info("Generating asset metadata.")

    data = data.groupby(by="symbol").agg({"date": ["min", "max"]})
    data.reset_index(inplace=True)
    data["start_date"] = data.date[np.min.__name__]
    data["end_date"] = data.date[np.max.__name__]
    del data["date"]
    data.columns = data.columns.get_level_values(0)

    data["exchange"] = "LIME"
    data["auto_close_date"] = data["end_date"].values + pd.Timedelta(days=1)
    return data


def parse_pricing_and_vol(data: pd.DataFrame, sessions: pd.IndexSlice, symbol_map: pd.Series):
    for asset_id, symbol in symbol_map.items():
        asset_data = (
            data.xs(symbol, level=1).infer_objects(copy=False).fillna(0.0)
        )
        yield asset_id, asset_data


def create_equities_bundle(
        bundle_name: str,
        period: Period,
        fundamental_data_list: set[str],
        symbol_list: list[str] = None,
):
    def ingest(
            environ: _Environ,
            historical_market_data_provider: AbstractHistoricalMarketDataProvider,
            fundamental_data_provider: AbstractFundamentalDataProvider,
            asset_db_writer: AssetDBWriter,
            minute_bar_writer: BcolzMinuteBarWriter,
            daily_bar_writer: BcolzDataBundle,
            fundamental_data_writer: BcolzDataBundle,
            adjustment_writer: SQLiteAdjustmentWriter,
            calendar: ExchangeCalendar,
            start_session: pd.Timestamp,
            end_session: pd.Timestamp,
            cache: dataframe_cache,
            show_progress: bool,
            market_data_fields: list[ColumnSpecification],
            fundamental_data_fields: list[ColumnSpecification],
            output_dir: str,
            **kwargs
    ):
        date_from = start_session.to_pydatetime().replace(tzinfo=datetime.timezone.utc)
        date_to = end_session.to_pydatetime().replace(tzinfo=datetime.timezone.utc)
        logger.info(f"Ingesting equities bundle {bundle_name} for period {start_session} - {end_session}")
        historical_data = historical_market_data_provider.get_historical_data_table(
            symbols=symbol_list,
            period=period,
            date_from=date_from,
            date_to=date_to,
            show_progress=show_progress)

        if len(historical_data) == 0:
            logger.warning(
                f"Data source {type(historical_market_data_provider)} didn't return any data for symbols {symbol_list}")
            return



        asset_metadata = gen_asset_metadata(data=historical_data[["symbol", "date"]], show_progress=show_progress)


        historical_data.set_index(["date", "symbol"], inplace=True)


        # historical_data = pd.concat([historical_data, fundamental_data], ignore_index=False, axis=1)
        # final_df = final_df[final_df.date.notnull()]


        exchanges = pd.DataFrame(
            data=[["LIME", "LIME", "US"]],
            columns=["exchange", "canonical_name", "country_code"],
        )

        asset_db_writer.write(equities=asset_metadata, exchanges=exchanges)


        symbol_map = asset_metadata.symbol
        sessions = calendar.sessions_in_range(start=start_session, end=end_session)




        fundamental_data = fundamental_data_provider.get_fundamental_data(symbols=symbol_list,
                                                                          period=period,
                                                                          date_from=date_from, date_to=date_to,
                                                                          fundamental_data_list=fundamental_data_list)

        fundamental_data_writer.write(
            data=parse_pricing_and_vol(data=fundamental_data, sessions=sessions, symbol_map=symbol_map),
            show_progress=show_progress,
            calendar=calendar,
            start_session=start_session,
            end_session=end_session,
            cols=fundamental_data_fields,
            validate_sessions=False
        )

        if period == Period.DAY:
            daily_bar_writer.write(
                data=parse_pricing_and_vol(data=historical_data, sessions=sessions, symbol_map=symbol_map),
                show_progress=show_progress,
                calendar=calendar,
                start_session=start_session,
                end_session=end_session,
                cols=market_data_fields,
                validate_sessions=False
            )
        else:
            raise Exception("Unsupported period.")


        # Write empty splits and divs - they are not present in API
        divs_splits = {
            "divs": pd.DataFrame(
                columns=[
                    "sid",
                    "amount",
                    "ex_date",
                    "record_date",
                    "declared_date",
                    "pay_date",
                ]
            ),
            "splits": pd.DataFrame(columns=["sid", "ratio", "effective_date"]),
        }
        adjustment_writer.write(
            splits=divs_splits["splits"], dividends=divs_splits["divs"]
        )
        logger.info(
            f"Ingesting equities bundle {bundle_name} for period {start_session} - {end_session} "
            f"and symbols {symbol_list} Completed"
        )

    return ingest


def register_lime_equities_bundle(
        bundle_name: str,
        start_session: datetime.datetime | None,
        end_session: datetime.datetime | None,
        symbol_list: list[str],
        period: Period,
        calendar_name: str,
        fundamental_data_list: set[str],
):
    if start_session and end_session:
        start_session, end_session = normalize_daily_start_end_session(
            calendar_name=calendar_name, start_session=start_session, end_session=end_session
        )
    register(
        name=bundle_name,
        f=create_equities_bundle(
            bundle_name=bundle_name,
            symbol_list=symbol_list,
            fundamental_data_list=fundamental_data_list,
            period=period
        ),
        start_session=start_session,
        end_session=end_session,
        calendar_name=calendar_name,
    )


register_calendar_alias("LIME", "NYSE")

# zipline.data.bcolz_daily_bars.BcolzDailyBarWriter = CustomClass



import datetime
from abc import abstractmethod
from lime_trader.models.market import Period


class AbstractLiveMarketDataProvider:

    @abstractmethod
    def fetch_live_data_table(
            self,
            symbols: list[str],
            period: Period,
            date_from: datetime.datetime,
            date_to: datetime.datetime,
            show_progress: bool):
        pass



import pandas as pd
from exchange_calendars import ExchangeCalendar

from ziplime.domain.column_specification import ColumnSpecification


class AbstractDataBundle:
    def write(
            self, data, calendar: ExchangeCalendar, start_session: pd.Timestamp, end_session: pd.Timestamp,
            cols: list[ColumnSpecification],
            validate_sessions: bool,
            assets=None, show_progress=False, invalid_data_behavior="warn",
            **kwargs
    ):
        pass



import datetime
import logging
import multiprocessing
import sys

import limexhub
import pandas as pd
from click import progressbar
from joblib import Parallel, delayed
from lime_trader.models.market import Period

from ziplime.data.abstract_historical_market_data_provider import AbstractHistoricalMarketDataProvider


def fetch_historical_limex_data_task(date_from: datetime.datetime,
                                     date_to: datetime.datetime,
                                     limex_api_key: str, symbol: str, period: Period):
    limex_client = limexhub.RestAPI(token=limex_api_key)
    timeframe = 3
    if period == Period.MINUTE:
        timeframe = 1
    elif period == Period.HOUR:
        timeframe = 2
    elif period == Period.DAY:
        timeframe = 3
    elif period == Period.WEEK:
        timeframe = 4
    elif period == Period.MONTH:
        timeframe = 5
    elif period == Period.QUARTER:
        timeframe = 6
    df = limex_client.candles(symbol=symbol,
                              from_date=date_from,
                              to_date=date_to,
                              timeframe=timeframe)
    if len(df) > 0:
        df = df.reset_index()
        df = df.rename(
            columns={"o": "open", "h": "high", "l": "low", "c": "close", "v": "volume", "Date": "date"})
        df["date"] = pd.to_datetime(df.date, utc=True)
        df = df.set_index('date', drop=False)
        df['dividend'] = 0
        df['split'] = 0

        df = df[df.date.notnull()]
        df["symbol"] = symbol
        return df
    return df


class LimexHubHistoricalMarketDataProvider(AbstractHistoricalMarketDataProvider):
    def __init__(self, limex_api_key: str):
        self._limex_api_key = limex_api_key
        self._logger = logging.getLogger(__name__)
        self._limex_client = limexhub.RestAPI(token=limex_api_key)

    def get_historical_data_table(self, symbols: list[str],
                                  period: Period,
                                  date_from: datetime.datetime,
                                  date_to: datetime.datetime,
                                  show_progress: bool,
                                  ):

        def fetch_historical(limex_api_key: str, symbol: str):
            try:
                result = fetch_historical_limex_data_task(date_from=date_from, date_to=date_to,
                                                          limex_api_key=limex_api_key, symbol=symbol, period=period)
                return result
            except Exception as e:
                logging.exception(
                    f"Exception fetching historical data for symbol {symbol}, date_from={date_from}, date_to={date_to}. Skipping."
                )
                return None

        total_days = (date_to - date_from).days
        final = pd.DataFrame()

        if show_progress:
            with progressbar(length=len(symbols) * total_days, label="Downloading historical data from LimexHub",
                             file=sys.stdout) as pbar:
                res = Parallel(n_jobs=multiprocessing.cpu_count() * 2, prefer="threads",
                               return_as="generator_unordered")(
                    delayed(fetch_historical)(self._limex_api_key, symbol) for symbol in symbols)
                for item in res:
                    pbar.update(total_days)
                    if item is None:
                        continue
                    final = pd.concat([final, item])
        else:
            res = Parallel(n_jobs=multiprocessing.cpu_count() * 2, prefer="threads", return_as="generator_unordered")(
                delayed(fetch_historical)(self._limex_api_key, symbol) for symbol in symbols)
            for item in res:
                if item is None:
                    continue
                final = pd.concat([final, item])
        final = final.sort_index()
        return final



import datetime
from abc import abstractmethod

from lime_trader.models.market import Period



class AbstractFundamentalDataProvider:

    @abstractmethod
    def get_fundamental_data(self,
                             symbols: list[str],
                             date_from: datetime.datetime,
                             date_to: datetime.datetime,
                             period: Period,
                             fundamental_data_list: set[str]
                             ):
        pass
